pub const STOPPING_ROUNDS: usize = 3;
pub const FREE_MEM_ALLOC_FACTOR: f32 = 0.9;
pub const N_NODES_ALLOC_LIMIT: usize = 3000;
pub const ITER_LIMIT: usize = 1000;
pub const GENERALIZATION_THRESHOLD: f32 = 0.99;
pub const MIN_COL_AMOUNT: usize = 40;
pub const HESSIAN_EPS: f32 = 1e-3;
pub const TIMEOUT_FACTOR: f32 = 0.95;
